/*
 * Copyright (C) 2020-2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/impl/lib/get_zone_data_context.hh"
#include "src/impl/lib/exception.hh"

#include "libpg/pg_ro_transaction.hh"
#include "libpg/query.hh"

#include "liblog/liblog.hh"

#include <boost/asio/ip/address.hpp>

#include <algorithm>
#include <future>
#include <thread>
#include <utility>

namespace boost {
namespace asio {
namespace ip {

inline address unwrap_from_psql_representation(const address&, const char* src)
{
#if 106600 <= BOOST_VERSION
    return make_address(src);
#else
    return address::from_string(src);//deprecated from the boost v1.66.0
#endif
}

}//namespace boost::asio::ip
}//namespace boost::asio
}//namespace boost

namespace Fred {
namespace Zone {
namespace Impl {
namespace Lib {

namespace {

decltype(auto) get_zone_ns(const LibPg::PgTransaction& tx, const ZoneName& zone)
{
    using Ns = Fqdn;
    using Addr = boost::asio::ip::address;
    const auto dbres = exec(
            tx,
            LibPg::make_query("SELECT") << LibPg::item<Ns>() << "n.fqdn"
                                        << LibPg::nullable_item<Addr>() << "UNNEST(CASE WHEN n.addrs = ARRAY[]::INET[] THEN '{NULL}'::INET[] ELSE n.addrs END) "
                              "FROM zone z "
                              "JOIN zone_ns n ON n.zone=z.id "
                              "WHERE z.fqdn=" << LibPg::parameter<ZoneName>().as_text(),
            {zone});
    std::map<AbsoluteFqdn, std::vector<Addr>> nameservers;
    std::for_each(dbres.begin(), dbres.end(), [&](auto row)
    {
        auto ns = AbsoluteFqdn{row.template get<Ns>()};
        if (row.template is_null<Addr>())
        {
            nameservers.emplace(std::move(ns), std::vector<Addr>{});
        }
        else
        {
            const auto ns_iter = nameservers.find(ns);
            if (ns_iter == nameservers.end())
            {
                nameservers.emplace(std::move(ns), std::vector<Addr>{row.template get_nullable<Addr>()});
            }
            else
            {
                ns_iter->second.emplace_back(row.template get_nullable<Addr>());
            }
        }
    });
    return nameservers;
}

ZoneData::Common get_common(const LibPg::PgTransaction& tx, const ZoneName& zone)
{
    using Ttl = ZoneData::Common::Ttl;
    using MName = StrongType<Fqdn, struct MNameTag_>;
    using EMail = StrongType<std::string, struct RNameTag_>;
    using RName = StrongType<EMail, struct RNameTag_>;
    using Serial = ZoneData::SoaRecord::RData::Serial;
    using Refresh = ZoneData::SoaRecord::RData::Refresh;
    using Retry = ZoneData::SoaRecord::RData::Retry;
    using Expire = ZoneData::SoaRecord::RData::Expire;
    using Minimum = ZoneData::SoaRecord::RData::Minimum;
    const auto dbres = exec(
            tx,
            LibPg::make_query("SELECT") << LibPg::item<Ttl>() << "zs.ttl"
                                        << LibPg::item<MName>() << "zs.ns_fqdn"
                                        << LibPg::item<RName>() << "zs.hostmaster"
                                        << LibPg::item<Serial>() << "EXTRACT('epoch' FROM DATE_TRUNC('second', NOW() AT TIME ZONE 'UTC'))"
                                        << LibPg::item<Refresh>() << "zs.refresh"
                                        << LibPg::item<Retry>() << "zs.update_retr"
                                        << LibPg::item<Expire>() << "zs.expiry"
                                        << LibPg::item<Minimum>() << "zs.minimum "
                              "FROM zone z "
                              "JOIN zone_soa zs ON zs.zone=z.id "
                              "WHERE z.fqdn=" << LibPg::parameter<ZoneName>().as_text(),
            {zone});
    if (dbres.empty())
    {
        const bool zone_exists = !exec(
                tx,
                LibPg::make_query("SELECT") << LibPg::item<int>() << "0 FROM zone WHERE fqdn=" << LibPg::parameter<ZoneName>().as_text(),
                {zone}).empty();
        if (zone_exists)
        {
            auto what = "no SOA record for zone " + *zone;
            throw_runtime_error([=]() { return what.c_str(); });
        }
        auto what = "zone " + *zone + " does not exist";
        throw_not_found([=]() { return what.c_str(); });
    }
    if (LibPg::RowIndex{1} < dbres.size())
    {
        throw_runtime_error([]() { return "too many SOA records"; });
    }
    const auto make_hostmaster = [&]()
    {
        auto hostmaster_email = **dbres.front().get<RName>();
        const auto idx_of_at = hostmaster_email.find_last_of('@');
        if (idx_of_at != std::string::npos)
        {
            std::string hostmaster = std::move(hostmaster_email);
            hostmaster[idx_of_at] = '.';
            return Fqdn{std::move(hostmaster)};
        }
        return Fqdn{std::move(hostmaster_email)};
    };
    return ZoneData::Common
    {
        dbres.front().get<Ttl>(),
        ZoneData::SoaRecord{
            zone,
            ZoneData::SoaRecord::RData{
                ZoneData::SoaRecord::RData::MName{make_absolute(*dbres.front().get<MName>())},
                ZoneData::SoaRecord::RData::RName{make_absolute(make_hostmaster())},
                dbres.front().get<Serial>(),
                dbres.front().get<Refresh>(),
                dbres.front().get<Retry>(),
                dbres.front().get<Expire>(),
                dbres.front().get<Minimum>()
            }},
        get_zone_ns(tx, zone)
    };
}

ZoneData::Context::DomainList get_domains(const LibPg::PgTransaction& tx, const ZoneName& zone)
{
    using KeysetId = ZoneData::Context::KeysetId;
    using NssetId = ZoneData::Context::NssetId;
    const auto dbres = exec(
            tx,
            LibPg::make_query("SELECT") << LibPg::item<Fqdn>() << "obr.name"
                                        << LibPg::nullable_item<KeysetId>() << "d.keyset"
                                        << LibPg::item<NssetId>() << "d.nsset "
                              "FROM domain d "
                              "JOIN object_registry obr ON obr.id=d.id "
                              "WHERE d.nsset IS NOT NULL AND "
                                    "NOT EXISTS(SELECT 0 FROM object_state WHERE object_id=d.id AND state_id=get_object_state_id('outzone') AND valid_to IS NULL) AND "
                                    "d.zone=(SELECT id FROM zone WHERE fqdn=" << LibPg::parameter<ZoneName>().as_text() << ")",
            {zone});
    auto domains = ZoneData::Context::DomainList{};
    domains.reserve(*dbres.size());
    std::transform(
            dbres.begin(),
            dbres.end(),
            back_inserter(domains),
            [](const auto& row)
            {
                ZoneData::Context::Domain domain{make_absolute(row.template get<Fqdn>())};
                if (!row.template is_null<KeysetId>())
                {
                    domain.keyset_id = row.template get_nullable<KeysetId>();
                }
                domain.nsset_id = row.template get<NssetId>();
                return domain;
            });
    return domains;
}

ZoneData::Context::KeysetList get_keysets(const LibPg::PgTransaction& tx, const ZoneName& zone)
{
    using KeysetId = ZoneData::Context::KeysetId;
    using Protocol = StrongType<std::uint16_t, struct ProtocolTag_>;
    using Algorithm = StrongType<std::uint16_t, struct AlgorithmTag_>;
    const auto dbres = exec(
            tx,
            LibPg::make_query("SELECT") << LibPg::item<KeysetId>() << "dnskey.keysetid"
                                        << LibPg::item<DnsKey::Flags>() << "dnskey.flags"
                                        << LibPg::item<Protocol>() << "dnskey.protocol"
                                        << LibPg::item<Algorithm>() << "dnskey.alg"
                                        << LibPg::item<DnsKey::PublicKey>() << "dnskey.key "
                              "FROM domain d "
                              "JOIN dnskey ON dnskey.keysetid=d.keyset "
                              "WHERE d.nsset IS NOT NULL AND "
                                    "NOT EXISTS(SELECT 0 FROM object_state WHERE object_id=d.id AND state_id=get_object_state_id('outzone') AND valid_to IS NULL) AND "
                                    "d.zone=(SELECT id FROM zone WHERE fqdn=" << LibPg::parameter<ZoneName>().as_text() << ") "
                              "GROUP BY 1,2,3,4,5 "
                              "ORDER BY 1",
            {zone});
    auto keyset_list = ZoneData::Context::KeysetList{};
    keyset_list.reserve(*dbres.size());
    std::for_each(
            dbres.begin(),
            dbres.end(),
            [&](const auto& row)
            {
                Fred::Zone::Impl::Lib::DnsKey dns_key{row.template get<DnsKey::Flags>(),
                                                      DnsKey::Protocol{static_cast<std::uint8_t>(*row.template get<Protocol>())},
                                                      DnsKey::Algorithm{static_cast<std::uint8_t>(*row.template get<Algorithm>())},
                                                      row.template get<DnsKey::PublicKey>()};
                ZoneData::Context::Keyset keyset{row.template get<KeysetId>(),
                                                 std::move(dns_key)};
                keyset_list.emplace_back(std::move(keyset));
            });
    return keyset_list;
}

ZoneData::Context::NssetList get_nssets(const LibPg::PgTransaction& tx, const ZoneName& zone)
{
    using NssetId = ZoneData::Context::NssetId;
    using IpAddr = Fred::Zone::Impl::Lib::StrongType<boost::asio::ip::address, struct IpAddrTag_>;
    const auto dbres = exec(
            tx,
            LibPg::make_query("SELECT") << LibPg::item<NssetId>() << "h.nssetid"
                                        << LibPg::item<Fqdn>() << "h.fqdn"
                                        << LibPg::nullable_item<IpAddr>() << "him.ipaddr "
                              "FROM domain d "
                              "JOIN host h ON h.nssetid=d.nsset "
                              "LEFT JOIN host_ipaddr_map him ON him.hostid=h.id AND him.nssetid=h.nssetid "
                              "WHERE NOT EXISTS(SELECT 0 FROM object_state WHERE object_id=d.id AND state_id=get_object_state_id('outzone') AND valid_to IS NULL) AND "
                                    "d.zone=(SELECT id FROM zone WHERE fqdn=" << LibPg::parameter<ZoneName>().as_text() << ") "
                              "GROUP BY 1,2,3 "
                              "ORDER BY 1,2",
            {zone});
    constexpr auto invalid_nsset_id = NssetId{-1};
    const auto number_of_nssets = [&]()
    {
        auto last_nsset_id = invalid_nsset_id;
        auto number_of_unique_nsset_id = std::int64_t{0};
        std::for_each(
                dbres.begin(),
                dbres.end(),
                [&](const auto& row)
                {
                    const auto nsset_id = row.template get<NssetId>();
                    if (nsset_id != last_nsset_id)
                    {
                        ++number_of_unique_nsset_id;
                        last_nsset_id = nsset_id;
                    }
                });
        return number_of_unique_nsset_id;
    }();
    auto nsset_list = ZoneData::Context::NssetList{};
    nsset_list.reserve(number_of_nssets);
    ZoneData::Context::Nsset top{invalid_nsset_id, {}};
    std::for_each(
            dbres.begin(),
            dbres.end(),
            [&](const auto& row)
            {
                const auto nsset_id = row.template get<NssetId>();
                if (nsset_id != top.nsset_id)
                {
                    if (top.nsset_id != invalid_nsset_id)
                    {
                        nsset_list.push_back(top);
                        top.nameservers.clear();
                    }
                    top.nsset_id = nsset_id;
                }
                auto fqdn = make_absolute(row.template get<Fqdn>());
                if (row.template is_null<IpAddr>())
                {
                    top.nameservers.insert(ZoneData::Context::Nsset::Nameservers::value_type{std::move(fqdn), {}});
                }
                else
                {
                    const auto current_nameserver_iter = top.nameservers.find(fqdn);
                    const bool has_fqdn = current_nameserver_iter != top.nameservers.end();
                    auto ip_addr = *row.template get_nullable<IpAddr>();
                    if (has_fqdn)
                    {
                        current_nameserver_iter->second.emplace_back(std::move(ip_addr));
                    }
                    else
                    {
                        top.nameservers.insert(ZoneData::Context::Nsset::Nameservers::value_type{std::move(fqdn), {std::move(ip_addr)}});
                    }
                }
            });
    if (top.nsset_id != invalid_nsset_id)
    {
        nsset_list.emplace_back(std::move(top));
    }
    return nsset_list;
}

template <typename T = void>
struct DataFromThread
{
    DataFromThread()
        : promise{},
          future{promise.get_future()}
    { }
    std::promise<T> promise;
    std::future<T> future;
};

template <typename T>
void get_zone_entity(
        LibPg::PgConnection connection,
        std::function<LibPg::PgRoTransaction(LibPg::PgConnection, std::promise<void>)> make_transaction,
        const ZoneName& zone,
        std::function<T(const LibPg::PgTransaction&, const ZoneName&)> get_fnc,
        std::promise<void> tx_start,
        std::promise<T> finished) noexcept
{
    try
    {
        try
        {
            auto tx = make_transaction(std::move(connection), std::move(tx_start));
            finished.set_value_at_thread_exit(get_fnc(tx, zone));
            commit(std::move(tx));
            return;
        }
        catch (const std::exception& e)
        {
            finished.set_exception_at_thread_exit(std::current_exception());
            LIBLOG_ERROR("std::exception caught: {}", e.what());
        }
        catch (...)
        {
            finished.set_exception_at_thread_exit(std::current_exception());
            LIBLOG_ERROR("unknown exception caught");
        }
    }
    catch (...) { }
}

}//namespace Fred::Zone::Impl::Lib::{anonymous}

ZoneData::Context get_zone_data_context(std::array<LibPg::PgConnection, 3> connections, const ZoneName& zone)
{
    try
    {
        auto main_tx = LibPg::PgRoTransaction{std::move(connections[0]), LibPg::PgTransaction::serializable_deferrable};
        const auto main_tx_snapshot_id = [](const LibPg::PgTransaction& tx)
        {
            static const auto select_snapshot_id = LibPg::make_query("SELECT") << LibPg::item<std::string>() << "pg_export_snapshot()";
            return exec(tx, select_snapshot_id).front().get<std::string>();
        }(main_tx);
        const auto clone_main_tx = [&](LibPg::PgConnection connection, std::promise<void> tx_start)
        {
            try
            {
                auto cloned_tx = LibPg::PgRoTransaction{std::move(connection), LibPg::PgTransaction::serializable};
                const auto use_snapshot = LibPg::make_query("SET TRANSACTION SNAPSHOT '" + main_tx_snapshot_id + "'");
                exec(cloned_tx, use_snapshot);
                tx_start.set_value();
                return cloned_tx;
            }
            catch (...)
            {
                tx_start.set_exception(std::current_exception());
                throw;
            }
        };

        ZoneData::Context zone_data_ctx{get_common(main_tx, zone)};
        auto get_domains_finished = DataFromThread<ZoneData::Context::DomainList>{};
        auto get_domains_tx_start = DataFromThread<>{};
        auto get_nssets_finished = DataFromThread<ZoneData::Context::NssetList>{};
        auto get_nssets_tx_start = DataFromThread<>{};
        auto get_domains_thread = std::thread{get_zone_entity<ZoneData::Context::DomainList>,
                                              std::move(connections[1]),
                                              clone_main_tx,
                                              std::cref(zone),
                                              get_domains,
                                              std::move(get_domains_tx_start.promise),
                                              std::move(get_domains_finished.promise)};
        auto get_nssets_thread = std::thread{get_zone_entity<ZoneData::Context::NssetList>,
                                              std::move(connections[2]),
                                              clone_main_tx,
                                              std::cref(zone),
                                              get_nssets,
                                              std::move(get_nssets_tx_start.promise),
                                              std::move(get_nssets_finished.promise)};
        zone_data_ctx.keysets = get_keysets(main_tx, zone);
        get_domains_tx_start.future.get();
        get_nssets_tx_start.future.get();
        commit(std::move(main_tx));
        zone_data_ctx.domains = get_domains_finished.future.get();
        zone_data_ctx.nssets = get_nssets_finished.future.get();
        get_domains_thread.join();
        get_nssets_thread.join();
        return zone_data_ctx;
    }
    catch (const NotFound& e)
    {
        LIBLOG_INFO("NotFound: {}", e.what());
        throw;
    }
    catch (const std::exception& e)
    {
        LIBLOG_ERROR("std::exception caught: {}", e.what());
        throw;
    }
    catch (...)
    {
        LIBLOG_ERROR("unknown exception caught");
        throw;
    }
}

}//namespace Fred::Zone::Impl::Lib
}//namespace Fred::Zone::Impl
}//namespace Fred::Zone
}//namespace Fred
